<?php

function _votingapi_bayesian_defined_functions() {
  return array(
    'bayesian' => array(
      'name' => t('Bayesian Average'),
      'description' => t('The Bayesian average from the votes.'),
      'module' => 'bayesian',
      'configuration' => array(
        'function_c' => array(
          'total_sum',
          'avg_vote_count',
          'num_distinct_item',
          'vote_count',
          'vote_count_by_item',
          'avg',
          'avg_by_item',
          'min',
          'max',
          'std',
          'std_by_item',
          'variance',
          'variance_by_item',
          'item_median',
          'items_median',
          'zero'
        ),
        'function_m' => array(
          'total_sum',
          'avg_vote_count',
          'num_distinct_item',
          'vote_count',
          'vote_count_by_item',
          'avg',
          'avg_by_item',
          'min',
          'max',
          'std',
          'std_by_item',
          'variance',
          'variance_by_item',
          'item_median',
          'items_median',
          'zero'
        )
      ),
      'default config' => array(
        'tags' => array(),
        'function_c' => 'avg_vote_count',
        'function_m' => 'items_median'
      )
    ),
    'weighted_arithmetic_mean' => array(
      'name' => t('Weighted arithmetic mean'),
      'description' => t('The Weighted arithmetic average from the votes.'),
      'module' => 'bayesian',
      'configuration' => array(
        'function_c' => array(
          'avg_vote_count',
        ),
        'function_m' => array(
          'avg',
        )
      ),
      'default config' => array(
        'tags' => array(),
        'function_c' => 'avg_vote_count',
        'function_m' => 'avg'
      )
    )
  );
}

function _votingapi_bayesian_get_vote_count_by_item($entity_type, $tag = 'vote', $entity_id = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');

  if (is_numeric($entity_id)) {
    $query->condition('entity_id', $entity_id, '=');
  }

  $query->addExpression('COUNT(vote_id)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_vote_count($entity_type, $tag = 'vote', $entity_id = NULL) {
  return _votingapi_bayesian_get_vote_count_by_item($entity_type, $tag, NULL);
}

function _votingapi_bayesian_get_total_sum($entity_type, $tag = 'vote') {
  $query = db_query('SELECT SUM(count) FROM (SELECT COUNT(vote_id)*value AS count FROM votingapi_vote WHERE entity_type = :type AND tag = :tag GROUP BY entity_id) AS SUM', array(':type' => $entity_type, ':tag' => $tag));
  return $query->fetchfield(0);
}

function _votingapi_bayesian_get_avg_by_item($entity_type, $tag = 'vote', $entity_id = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');

  if (is_numeric($entity_id)) {
    $query->condition('entity_id', $entity_id, '=');
  }

  $query->addExpression('AVG(value)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_zero($entity_type, $tag = 'vote', $entity_id = NULL) {
  return 0;
}

function _votingapi_bayesian_get_avg($entity_type, $tag = 'vote', $entity_id = NULL) {
  return _votingapi_bayesian_get_avg_by_item($entity_type, $tag, NULL);
}

function _votingapi_bayesian_get_avg_vote_count($entity_type, $tag = 'vote', $empty = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');
  $query->addExpression('COUNT(entity_id)/COUNT(DISTINCT entity_id)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_num_distinct_item($entity_type, $tag = 'vote', $empty = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');
  $query->addExpression('COUNT(entity_id)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_min($entity_type, $tag = 'vote', $entity_id = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');

  if (is_numeric($entity_id)) {
    $query->condition('entity_id', $entity_id, '=');
  }

  $query->addExpression('MIN(value)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_max($entity_type, $tag = 'vote', $entity_id = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');

  if (is_numeric($entity_id)) {
    $query->condition('entity_id', $entity_id, '=');
  }

  $query->addExpression('MAX(value)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_item_median($entity_type, $tag = 'vote', $entity_id) {
  $query = db_query('SELECT x.value from votingapi_vote x, votingapi_vote y WHERE x.entity_type = :type AND x.tag = :tag AND x.entity_id = :id GROUP BY x.value HAVING SUM(SIGN(1-SIGN(y.value-x.value)))/COUNT(*) > .5 LIMIT 1;', array(':type' => $entity_type, ':tag' => $tag, 'id' => $entity_id));
  return $query->fetchfield(0);
}

function _votingapi_bayesian_get_items_median($entity_type, $tag = 'vote', $empty) {
  $query = db_query('SELECT x.value from votingapi_vote x, votingapi_vote y WHERE x.entity_type = :type AND x.tag = :tag GROUP BY x.value HAVING SUM(SIGN(1-SIGN(y.value-x.value)))/COUNT(*) > .5 LIMIT 1;', array(':type' => $entity_type, ':tag' => $tag));
  return $query->fetchfield(0);
}

function _votingapi_bayesian_get_std_by_item($entity_type, $tag = 'vote', $entity_id = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');

  if (is_numeric($entity_id)) {
    $query->condition('entity_id', $entity_id, '=');
  }

  $query->addExpression('STD(value)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_std($entity_type, $tag = 'vote', $entity_id = NULL) {
  return _votingapi_bayesian_get_std_by_item($entity_type, $tag, NULL);
}

function _votingapi_bayesian_get_variance_by_item($entity_type, $tag = 'vote', $entity_id = NULL) {
  $query = db_select('votingapi_vote', 'v')
    ->condition('tag', $tag, '=')
    ->condition('entity_type', $entity_type, '=');

  if (is_numeric($entity_id)) {
    $query->condition('entity_id', $entity_id, '=');
  }

  $query->addExpression('VARIANCE(value)');

  return $query->execute()->fetchfield(0);
}

function _votingapi_bayesian_get_variance($entity_type, $tag = 'vote', $entity_id = NULL) {
  return _votingapi_bayesian_get_variance_by_item($entity_type, $tag, NULL);
}

function _votingapi_bayesian_get_weighted_arithmetic_mean($entity_type, $entity_id, $config) {
  $item_num_vote = _votingapi_bayesian_get_vote_count($entity_type, $config['tag'], $entity_id);
  $item_avg_rating = _votingapi_bayesian_get_avg_by_item($entity_type, $config['tag'], $entity_id);

  $function_c = '_votingapi_bayesian_get_' . $config['function_c'];
  $function_m = '_votingapi_bayesian_get_' . $config['function_m'];

  $c = $function_c($entity_type, $config['tag'], $entity_id);
  $m = $function_m($entity_type, $config['tag'], $entity_id);

  $result = 0;

  if ( 0 != ($c + $item_num_vote)) {
    $result = (($c*$m + ($item_num_vote * $item_avg_rating)) / ($c + $item_num_vote));
  }

  return $result;
}

function _votingapi_bayesian_get_bayesian($entity_type, $entity_id, $config) {
  /**
   * Bayesian formula
   *
   * (C*m) + (ITEM_NUM_VOTE * ITEM_AVG_RATING)
   * ---------------------------------------------
   *         (C + ITEM_NUM_VOTE)
   *
   * See: http://goo.gl/EQ0D9M
   *
   */

  $function_c = '_votingapi_bayesian_get_' . $config['function_c'];
  $function_m = '_votingapi_bayesian_get_' . $config['function_m'];

  $item_num_vote   = _votingapi_bayesian_get_vote_count($entity_type, $config['tag'], $entity_id);
  $item_avg_rating = _votingapi_bayesian_get_avg_by_item($entity_type, $config['tag'], $entity_id);

  /*
   * C is assigned a value that is proportional to the typical data set size.
   * The value is larger when the expected variation between data sets (within the larger population) is small.
   * It is smaller, when the data sets are expected to vary substantially from one another.
   */
  $c = $function_c($entity_type, $config['tag'], $entity_id);

  /*
   * In cases where the averages' relative values are the only result of importance, m can be replaced with zero.
   */
  $m = $function_m($entity_type, $config['tag'], $entity_id);

  $result = 0;

  if ( 0 != ($c + $item_num_vote)) {
    $result = (($c*$m + ($item_num_vote * $item_avg_rating)) / ($c + $item_num_vote));
  }

  return $result;
}

function _votingapi_bayesian_regenerate($function, $entity_type, $config) {
  $function_name = '_votingapi_bayesian_get_' . $function;

  $query = db_select('votingapi_cache', 'v')->distinct()
    ->fields('v', array('entity_id'))
    ->condition('tag', $config['tag'], '=')
    ->condition('entity_type', $entity_type, '=');
  $result = $query->execute();

  while($record = $result->fetchObject()) {
    db_update('votingapi_cache')
      ->fields(array(
        'value' => $function_name($entity_type, $record->entity_id, $config)
      ))
      ->condition('function', $function, '=')
      ->condition('entity_type', $entity_type, '=')
      ->condition('tag', $config['tag'], '=')
      ->condition('entity_id', $record->entity_id, '=')
      ->execute();
  }
}
